########################################################################################################################
# Module: tests/test_transport.py
# Description: Tests for SVGD and SMC samplers.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################


import unittest
from typing import Tuple

from jax import numpy as jnp, random
import numpy.testing as npt

from mocat.src.core import cdict
from mocat.src.sample import run
from mocat.src.scenarios.twodim import toy_examples
from mocat.src.transport.smc import MetropolisedSMCSampler
from mocat.src.transport.svgd import SVGD
from mocat.src.kernels import median_bandwidth_update, mean_bandwidth_update
from mocat.src.mcmc.standard_mcmc import RandomWalk, Overdamped


class TestCorrelatedGaussian(unittest.TestCase):
    scenario_cov = jnp.array([[1., 0.9], [0.9, 2.]])
    scenario = toy_examples.Gaussian(covariance=scenario_cov)
    n = int(1e4)

    def _test_mean(self,
                   sample: cdict):
        if sample.value.ndim == 3:
            val = jnp.concatenate(sample.value)
        else:
            val = sample.value
        samp_mean = val.mean(axis=0)
        npt.assert_array_almost_equal(samp_mean, jnp.zeros(2), decimal=1)

    def _test_cov(self,
                  sample: cdict):
        if sample.value.ndim == 3:
            val = jnp.concatenate(sample.value)
        else:
            val = sample.value
        samp_cov = jnp.cov(val.T)
        npt.assert_array_almost_equal(samp_cov, self.scenario_cov, decimal=1)


class TestSVGD(TestCorrelatedGaussian):
    n = int(1e2)
    n_iter = int(1e3)

    def test_fixed_kernel_params(self):
        class SVGD_fixed(SVGD):
            def adapt(self,
                      ensemble_state: cdict,
                      ensemble_extra: cdict) -> Tuple[cdict, cdict]:
                return ensemble_state, ensemble_extra

        sample = run(self.scenario, SVGD_fixed(max_iter=self.n_iter, stepsize=0.8),
                     n=self.n,
                     random_key=random.PRNGKey(0))
        self._test_mean(sample[-1])
        self._test_cov(sample[-1])

    def test_median_update(self):
        class SVGD_median(SVGD):
            def adapt(self,
                      ensemble_state: cdict,
                      ensemble_extra: cdict) -> Tuple[cdict, cdict]:
                ensemble_extra.parameters.kernel_params.bandwidth = median_bandwidth_update(ensemble_state.value)
                ensemble_state.kernel_params = ensemble_extra.parameters.kernel_params
                return ensemble_state, ensemble_extra

        sample = run(self.scenario, SVGD_median(max_iter=self.n_iter, stepsize=1.0),
                     n=self.n,
                     random_key=random.PRNGKey(0))

        self._test_mean(sample[-1])
        self._test_cov(sample[-1])

    def test_mean_update(self):
        class SVGD_mean(SVGD):
            def adapt(self,
                      ensemble_state: cdict,
                      ensemble_extra: cdict) -> Tuple[cdict, cdict]:
                ensemble_extra.parameters.kernel_params.bandwidth = mean_bandwidth_update(ensemble_state.value)
                ensemble_state.kernel_params = ensemble_extra.parameters.kernel_params
                return ensemble_state, ensemble_extra

        sample = run(self.scenario, SVGD_mean(max_iter=self.n_iter, stepsize=1.0),
                     n=self.n,
                     random_key=random.PRNGKey(0))

        self._test_mean(sample)
        self._test_cov(sample)

    def test_callable_stepsize(self):
        sample = run(self.scenario, SVGD(max_iter=self.n_iter, stepsize=lambda i: i ** -0.5),
                     n=self.n,
                     random_key=random.PRNGKey(0))

        self._test_mean(sample)
        self._test_cov(sample)

    def test_ensemble_minibatch(self):
        sample = run(self.scenario, SVGD(max_iter=self.n_iter, stepsize=1.0, ensemble_batchsize=100),
                     n=1000,
                     random_key=random.PRNGKey(0))

        self._test_mean(sample)
        self._test_cov(sample)


class TestMetropolisedSMC(TestCorrelatedGaussian):
    preschedule = jnp.arange(0., 1.1, 0.1)

    def resample_final(self,
                       sample: cdict) -> cdict:
        unweighted_vals = sample.value[-1, random.categorical(random.PRNGKey(1),
                                                              logits=sample.log_weight[-1], shape=(self.n,))]
        unweighted_sample = cdict(value=unweighted_vals)
        return unweighted_sample

    def test_tempered_preschedule_RW(self):
        sample = run(self.scenario, MetropolisedSMCSampler(RandomWalk(stepsize=1.0)),
                     self.n,
                     random_key=random.PRNGKey(0),
                     temperature_schedule=self.preschedule)
        unweighted_sample = self.resample_final(sample)
        self._test_mean(unweighted_sample)
        self._test_cov(unweighted_sample)
        npt.assert_array_equal(sample.temperature, self.preschedule[1:])

    def test_tempered_preschedule_OD(self):
        sample = run(self.scenario, MetropolisedSMCSampler(Overdamped(stepsize=1.0)),
                     self.n,
                     random_key=random.PRNGKey(0),
                     temperature_schedule=self.preschedule)
        unweighted_sample = self.resample_final(sample)
        self._test_mean(unweighted_sample)
        self._test_cov(unweighted_sample)
        npt.assert_array_equal(sample.temperature, self.preschedule[1:])

    def test_tempered_adaptive_RW(self):
        sample = run(self.scenario, MetropolisedSMCSampler(RandomWalk(stepsize=1.0)),
                     self.n,
                     random_key=random.PRNGKey(0))
        unweighted_sample = self.resample_final(sample)
        self._test_mean(unweighted_sample)
        self._test_cov(unweighted_sample)

    def test_tempered_adaptive_OD(self):
        sample = run(self.scenario, MetropolisedSMCSampler(Overdamped(stepsize=1.0)),
                     self.n,
                     random_key=random.PRNGKey(0))
        unweighted_sample = self.resample_final(sample)
        self._test_mean(unweighted_sample)
        self._test_cov(unweighted_sample)


if __name__ == '__main__':
    unittest.main()




from mocat.src.ssm.scenarios import scenarios

from mocat.src.ssm.ssm import StateSpaceModel

from mocat.src.ssm.filters import ParticleFilter
from mocat.src.ssm.filters import BootstrapFilter
from mocat.src.ssm.filters import initiate_particles
from mocat.src.ssm.filters import resample_particles
from mocat.src.ssm.filters import propagate_particle_filter
from mocat.src.ssm.filters import run_particle_filter_for_marginals

from mocat.src.ssm.backward import backward_simulation
from mocat.src.ssm.backward import forward_filtering_backward_simulation

from mocat.src.ssm.linear_gaussian.linear_gaussian import LinearGaussian
from mocat.src.ssm.linear_gaussian.linear_gaussian import TimeHomogenousLinearGaussian

from mocat.src.ssm.nonlinear_gaussian import NonLinearGaussian
from mocat.src.ssm.nonlinear_gaussian import OptimalNonLinearGaussiajnparticleFilter
from mocat.src.ssm.nonlinear_gaussian import EnsembleKalmanFilter





########################################################################################################################
# Module: abc/smc.py
# Description: SMC sampler for ABC problems.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from typing import Union, Tuple, Type
from inspect import isclass

from jax import numpy as jnp, random, vmap
from jax.lax import cond, scan

from mocat.src.abc.abc import ABCScenario, ABCSampler
from mocat.src.abc.mcmc import ABCMCMCSampler, RandomWalkABC
from mocat.src.core import cdict, is_implemented
from mocat.src.transport.smc import SMCSampler
from mocat.src.transport.sampler import TransportSampler
from mocat.src.mcmc.sampler import Correction, check_correction


class ABCSMCSampler(ABCSampler, SMCSampler):
    name = "ABC SMC"

    def __init__(self,
                 threshold_schedule: Union[None, jnp.ndarray] = None,
                 max_iter: int = int(1e4),
                 **kwargs):
        self.max_iter = max_iter
        self.threshold_schedule = threshold_schedule
        super().__init__(**kwargs)

    def __setattr__(self, key, value):
        super().__setattr__(key, value)
        if key == 'threshold_schedule':
            if value is None:
                self.next_threshold = self.next_threshold_adaptive
                self.max_iter = int(1e4)
            else:
                self.next_threshold = lambda state, extra: self.threshold_schedule[extra.iter[0]]
                self.max_iter = len(value)

    def startup(self,
                abc_scenario: ABCScenario,
                n: int,
                random_key: jnp.ndarray = None,
                initial_state: cdict = None,
                initial_extra: cdict = None,
                **kwargs) -> Tuple[cdict, cdict]:

        initial_state, initial_extra = TransportSampler.startup(self, abc_scenario, n, random_key,
                                                                initial_state, initial_extra, **kwargs)

        n = len(initial_state.value)
        if not hasattr(initial_state, 'prior_potential') and is_implemented(abc_scenario.prior_potential):
            double_random_keys = random.split(initial_extra.random_key[0], 2*n)
            initial_extra.random_key = double_random_keys[:n]
            initial_state.prior_potential = vmap(abc_scenario.prior_potential)(initial_state.value,
                                                                               double_random_keys[n:])

        if not hasattr(initial_state, 'log_weight'):
            initial_state.log_weight = jnp.zeros(n)

        if not hasattr(initial_state, 'simulated_data'):
            double_random_keys = random.split(initial_extra.random_key[0], 2*n)
            initial_extra.random_key = double_random_keys[:n]
            initial_extra.simulated_data = vmap(abc_scenario.likelihood_sample)(initial_state.value,
                                                                                double_random_keys[n:])

        if not hasattr(initial_state, 'distance'):
            initial_state.distance = vmap(abc_scenario.distance_function)(initial_extra.simulated_data)

        initial_extra.parameters = vmap(lambda _: initial_extra.parameters)(jnp.arange(n))

        if not hasattr(initial_state, 'threshold'):
            if self.threshold_schedule is None:
                initial_state.threshold = jnp.zeros(n) + jnp.inf
            else:
                initial_state.threshold = jnp.zeros(n) + self.threshold_schedule[0]

        return initial_state, initial_extra

    def next_threshold_adaptive(self,
                                state: cdict,
                                extra: cdict) -> float:
        raise AttributeError(f'{self.name} next_threshold_adaptive not initiated')

    def forward_proposal(self,
                         abc_scenario: ABCScenario,
                         previous_state: cdict,
                         previous_extra: cdict) -> Tuple[cdict, cdict]:
        raise AttributeError(f'{self.name} forward_proposal not initiated')

    def clean_chain(self,
                    abc_scenario: ABCScenario,
                    chain_ensemble_state: cdict) -> cdict:
        chain_ensemble_state.threshold = chain_ensemble_state.threshold[:, 0]
        return chain_ensemble_state


def adapt_stepsize_scaled_diag_cov(ensemble_state: cdict,
                                   ensemble_extra: cdict) -> Tuple[cdict, cdict]:
    n, d = ensemble_state.value.shape
    stepsize = vmap(jnp.cov, (1,))(ensemble_state.value) / d * 2.38 ** 2
    ensemble_extra.parameters.stepsize = jnp.repeat(stepsize[jnp.newaxis], n, axis=0)
    return ensemble_state, ensemble_extra


class MetropolisedABCSMCSampler(ABCSMCSampler):

    def __init__(self,
                 mcmc_sampler: Union[ABCMCMCSampler, Type[ABCMCMCSampler]] = None,
                 mcmc_correction: Union[Correction, Type[Correction], str] = 'sampler_default',
                 mcmc_steps: int = 20,
                 threshold_schedule: Union[None, jnp.ndarray] = None,
                 max_iter: int = int(1e4),
                 threshold_quantile_retain: float = 0.75,
                 **kwargs):
        super().__init__(max_iter=max_iter, threshold_schedule=threshold_schedule, **kwargs)
        if mcmc_sampler is None:
            mcmc_sampler = RandomWalkABC()
            self.adapt = adapt_stepsize_scaled_diag_cov

        if isclass(mcmc_sampler):
            mcmc_sampler = mcmc_sampler()
        self.mcmc_sampler = mcmc_sampler
        if mcmc_correction != 'sampler_default':
            self.mcmc_sampler.correction = mcmc_correction
        self.parameters.mcmc_steps = mcmc_steps
        self.parameters.threshold_quantile_retain = threshold_quantile_retain

    def startup(self,
                abc_scenario: ABCScenario,
                n: int,
                random_key: jnp.ndarray = None,
                initial_state: cdict = None,
                initial_extra: cdict = None,
                **kwargs) -> Tuple[cdict, cdict]:

        self.mcmc_sampler.correction = check_correction(self.mcmc_sampler.correction)

        initial_state, initial_extra = super().startup(abc_scenario, n, random_key, initial_state, initial_extra,
                                                       **kwargs)

        initial_state, initial_extra = vmap(
            lambda state, extra: self.mcmc_sampler.startup(abc_scenario,
                                                           n,
                                                           None,
                                                           state,
                                                           extra))(initial_state, initial_extra)

        initial_threshold = self.next_threshold(initial_state, initial_extra)
        initial_state.threshold = jnp.zeros(n) + initial_threshold
        initial_extra.parameters.threshold = initial_state.threshold
        if not hasattr(initial_state, 'log_weight'):
            initial_state.log_weight = jnp.zeros(n)
        initial_state.log_weight = jnp.where(initial_state.distance < initial_threshold,
                                            initial_state.log_weight, -jnp.inf)
        initial_state, initial_extra = self.adapt(initial_state, initial_extra)
        return initial_state, initial_extra

    def termination_criterion(self,
                              ensemble_state: cdict,
                              ensemble_extra: cdict) -> bool:
        return jnp.logical_or(ensemble_state.alpha.mean() <= 0.01,
                             ensemble_extra.iter[0] >= self.max_iter)

    def next_threshold_adaptive(self,
                                state: cdict,
                                extra: cdict) -> float:
        return jnp.quantile(state.distance, extra.parameters.threshold_quantile_retain[0])

    def log_weight(self,
                   previous_ensemble_state: cdict,
                   previous_ensemble_extra: cdict,
                   new_ensemble_state: cdict,
                   new_ensemble_extra: cdict) -> jnp.ndarray:
        return jnp.where(new_ensemble_state.distance < new_ensemble_extra.parameters.threshold, 0., -jnp.inf)

    def clean_mcmc_chain(self,
                         chain_state: cdict,
                         chain_extra: cdict) -> Tuple[cdict, cdict]:
        clean_state = chain_state[-1]
        clean_extra = chain_extra[-1]
        clean_state.alpha = chain_state.alpha.mean()
        clean_extra.parameters = chain_extra.parameters[-1]
        return clean_state, clean_extra

    def forward_proposal(self,
                         abc_scenario: ABCScenario,
                         state: cdict,
                         extra: cdict) -> Tuple[cdict, cdict]:

        def mcmc_kernel(previous_carry: Tuple[cdict, cdict],
                        _: None) -> Tuple[Tuple[cdict, cdict], Tuple[cdict, cdict]]:
            new_carry = self.mcmc_sampler.update(abc_scenario, *previous_carry)
            return new_carry, new_carry

        start_state, start_extra = self.mcmc_sampler.startup(abc_scenario,
                                                             extra.parameters.mcmc_steps,
                                                             None,
                                                             state,
                                                             extra)

        final_carry, chain = scan(mcmc_kernel,
                                  (start_state, start_extra),
                                  None,
                                  length=self.parameters.mcmc_steps)

        advanced_state, advanced_extra = self.clean_mcmc_chain(chain[0], chain[1])
        advanced_extra.iter = extra.iter

        return advanced_state, advanced_extra

    def update(self,
               abc_scenario: ABCScenario,
               ensemble_state: cdict,
               ensemble_extra: cdict) -> Tuple[cdict, cdict]:
        n = ensemble_state.value.shape[0]
        ensemble_extra.iter = ensemble_extra.iter + 1

        resample_bool = self.resample_criterion(ensemble_state, ensemble_extra)
        ensemble_state.log_weight = jnp.where(resample_bool, 0., ensemble_state.log_weight)

        resampled_ensemble_state, resampled_ensemble_extra \
            = cond(resample_bool,
                   lambda tup: self.resample(*tup),
                   lambda tup: tup,
                   (ensemble_state, ensemble_extra))

        advanced_state, advanced_extra = vmap(self.forward_proposal, in_axes=(None, 0, 0))(abc_scenario,
                                                                                           resampled_ensemble_state,
                                                                                           resampled_ensemble_extra)

        advanced_extra.iter = ensemble_extra.iter
        next_threshold = self.next_threshold(advanced_state, advanced_extra)
        advanced_state.threshold = jnp.ones(n) * next_threshold
        advanced_extra.parameters.threshold = advanced_state.threshold
        advanced_state.log_weight = self.log_weight(ensemble_state, ensemble_extra, advanced_state, advanced_extra)

        advanced_state, advanced_extra = self.adapt(advanced_state, advanced_extra)

        return advanced_state, advanced_extra




########################################################################################################################
# Module: twodim/vectorise.py
# Description: Two dimensional scenario class - vectorises potentials (that aren't already)
#              and automatically finds limits for plotting.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from typing import Callable

import jax.numpy as jnp
import matplotlib.pyplot as plt

from mocat.src.core import Scenario


def _flex_vectorise(func: Callable):
    try:
        # Check if already vectorised
        test_eval = func(jnp.zeros((3, 2)), None)
        if test_eval.shape != (3, ):
            raise TypeError
    except:
        # Attempt to vectorise
        def vec_function(arr, rk=None):
            if arr.ndim == 1 and arr.shape[-1] == 2:
                return func(arr, rk)
            elif arr.ndim == 2 and arr.shape[-1] == 2:
                return func(arr.transpose(), rk)
            elif arr.ndim == 3 and arr.shape[-1] == 2:
                return func(arr.transpose((2, 0, 1)), rk)
            else:
                raise TypeError('Array of incorrect dimension for vectorisation')

        test_eval = vec_function(jnp.zeros((3, 2)), None)
        if test_eval.shape != (3, ):
            raise TypeError('Scenario potential vectorised successfully but failed sanity check')

        return vec_function
    else:
        return func


def _generate_plot_grid(x_lim: list,
                        y_lim: list = None,
                        resolution: int = 100,
                        linspace: bool = False):
    if y_lim is None:
        y_lim = x_lim
    x_linspace = jnp.linspace(x_lim[0], x_lim[1], resolution)
    y_linspace = jnp.linspace(y_lim[0], y_lim[1], resolution)
    x_grid, y_grid = jnp.meshgrid(x_linspace, y_linspace)
    joint_grid = jnp.stack([x_grid, y_grid], axis=2)

    return [x_linspace, y_linspace, joint_grid] if linspace else joint_grid


def _find_first_non_zero_row(matrix: jnp.ndim,
                             direction: int = 1):
    if direction == 1:
        i = 0
        while not any(matrix[i]):
            i += 1
        return i
    elif direction == -1:
        i = len(matrix) - 1
        while not any(matrix[i]):
            i += -1
        return i


def auto_axes_lims(vec_dens: Callable,
                   xlim: float = 10.,
                   ylim: float = 10.):
    # Assumes ijnput is a vectorised function that goes to 0 in the tails

    # Initial evaluation grid
    ix, iy, grid = _generate_plot_grid([-xlim, xlim], [-ylim, ylim], resolution=100, linspace=True)
    z = vec_dens(grid)

    # Find mode
    max_z = jnp.max(z)
    if jnp.isnan(max_z):
        raise TypeError('nan found attempting auto_axes_lims, try giving manual xlim and ylim as kwargs')

    if max_z == 0.:
        return auto_axes_lims(vec_dens, xlim=xlim/1.5, ylim=ylim/1.5)

    # Area with probability mass
    z_keep = z > max_z / 10

    # Find bounds of area with probability mass
    xlim_new = jnp.array([ix[_find_first_non_zero_row(z_keep.T, direction)] for direction in [1, -1]])
    ylim_new = jnp.array([iy[_find_first_non_zero_row(z_keep, direction)] for direction in [1, -1]])

    if xlim in jnp.abs(xlim_new) or ylim in jnp.abs(ylim_new):
        return auto_axes_lims(vec_dens,
                              xlim=2*xlim if xlim in jnp.abs(xlim_new) else xlim,
                              ylim=2*ylim if ylim in jnp.abs(ylim_new) else ylim)

    # Expand
    expansion = 0.05
    xlim += (xlim_new[1] - xlim_new[0]) * expansion * jnp.array([-1, 1])
    ylim += (ylim_new[1] - ylim_new[0]) * expansion * jnp.array([-1, 1])

    return tuple(xlim_new), tuple(ylim_new)


def _plot_densf(ax, x, y, z, **kwargs):
    return ax.contourf(x, y, z, **kwargs)


class TwoDimToyScenario(Scenario):

    dim = 2
    xlim = None
    ylim = None
    vec_dens = None
    vec_potential = None
    plot_resolution = 1000

    def prior_potential(self,
                        x: jnp.ndarray,
                        random_key: jnp.ndarray = None) -> float:
        return 0.

    def _vectorise(self):
        self.vec_potential = _flex_vectorise(self.potential)
        self.vec_dens = lambda x, rk = None: jnp.exp(-self.vec_potential(x, rk))

    def auto_axes_lims(self, vectorise=True):
        if vectorise:
            self._vectorise()
        self.xlim , self.ylim = auto_axes_lims(self.vec_dens)

    def plot(self, ax=None, xlim=None, ylim=None, recalc_axes=False, potential=False,
             cmap='Purples', **kwargs):
        self._vectorise()

        # Check plot bounds
        if ax is not None:
            xlim = ax.get_xlim()
            ylim = ax.get_ylim()
        elif xlim is None or ylim is None:
            # Generate plot dimensions if not already present
            if self.xlim is None or recalc_axes:
                self.auto_axes_lims(vectorise=False)
            xlim = self.xlim if xlim is None else xlim
            ylim = self.ylim if ylim is None else ylim

        # Generate plot grid
        x_linsp, y_linsp, grid = _generate_plot_grid(xlim, ylim, self.plot_resolution, True)

        # Determine plot function
        plot_func = self.vec_potential if potential else self.vec_dens

        if ax is None:
            # Plot contours
            fig, ax = plt.subplots()
            _plot_densf(ax, x_linsp, y_linsp, plot_func(grid), cmap=cmap, **kwargs)
            return fig, ax
        else:
            return _plot_densf(ax, x_linsp, y_linsp, plot_func(grid), cmap=cmap, **kwargs)



########################################################################################################################
# Module: ssm/filters.py
# Description: Particle filtering inc bootstrap filter.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from functools import partial
from typing import Union, Tuple

import jax.numpy as jnp
from jax import random, vmap, jit
from jax.lax import cond, scan
from jax.ops import index_update

from mocat.src.core import cdict
from mocat.src.ssm.ssm import StateSpaceModel
from mocat.src.metrics import ess_log_weight


class ParticleFilter:
    name = 'Particle Filter'

    def __init__(self,
                 name: str = None):
        if name is not None:
            self.name = name

    def __repr__(self):
        return f"mocat.ParticleFilter.{self.__class__.__name__}"

    def startup(self,
                ssm_scenario: StateSpaceModel):
        pass

    def initial_potential(self,
                          ssm_scenario: StateSpaceModel,
                          x: jnp.ndarray,
                          y: jnp.ndarray,
                          t: float) -> Union[float, jnp.ndarray]:
        return ssm_scenario.initial_potential(x, t)

    def initial_sample(self,
                       ssm_scenario: StateSpaceModel,
                       y: jnp.ndarray,
                       t: float,
                       random_key: jnp.ndarray) -> jnp.ndarray:
        return ssm_scenario.initial_sample(t, random_key)

    def initial_log_weight(self,
                           ssm_scenario: StateSpaceModel,
                           x: jnp.ndarray,
                           y: jnp.ndarray,
                           t: float) -> Union[float, jnp.ndarray]:
        return -ssm_scenario.likelihood_potential(x, y, t)

    @partial(jit, static_argnums=(0, 1))
    def initial_sample_vectorised(self,
                                  ssm_scenario: StateSpaceModel,
                                  y: jnp.ndarray,
                                  t: float,
                                  random_keys: jnp.ndarray) -> jnp.ndarray:
        init_vals = vmap(self.initial_sample, (None, None, None, 0))(ssm_scenario, y, t, random_keys)
        return init_vals

    @partial(jit, static_argnums=(0, 1))
    def initial_sample_and_weight_vectorised(self,
                                             ssm_scenario: StateSpaceModel,
                                             y: jnp.ndarray,
                                             t: float,
                                             random_keys: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:
        init_vals = vmap(self.initial_sample, (None, None, None, 0))(ssm_scenario, y, t, random_keys)

        if y is not None:
            init_log_weight = vmap(self.initial_log_weight, (None, 0, None, None))(ssm_scenario, init_vals, y, t)
        else:
            init_log_weight = jnp.zeros(len(init_vals))

        return init_vals, init_log_weight

    def proposal_potential(self,
                           ssm_scenario: StateSpaceModel,
                           x_previous: jnp.ndarray,
                           t_previous: float,
                           x_new: jnp.ndarray,
                           y_new: jnp.ndarray,
                           t_new: float) -> Union[float, jnp.ndarray]:
        raise AttributeError(f'{self.__class__.__name__} transition_potential not implemented')

    def proposal_sample(self,
                        ssm_scenario: StateSpaceModel,
                        x_previous: jnp.ndarray,
                        t_previous: float,
                        y_new: jnp.ndarray,
                        t_new: float,
                        random_key: jnp.ndarray) -> jnp.ndarray:
        raise AttributeError(f'{self.__class__.__name__} proposal_sample not implemented')

    def intermediate_log_weight(self,
                                ssm_scenario: StateSpaceModel,
                                x_previous: jnp.ndarray,
                                t_previous: float,
                                x_new: jnp.ndarray,
                                y_new: jnp.ndarray,
                                t_new: float) -> Union[float, jnp.ndarray]:
        return - ssm_scenario.transition_potential(x_previous, t_previous,
                                                   x_new, t_new) \
               - ssm_scenario.likelihood_potential(x_new, y_new, t_new) \
               + self.proposal_potential(ssm_scenario, x_previous, t_previous, x_new, y_new, t_new)

    @partial(jit, static_argnums=(0, 1))
    def proposal_sample_vectorised(self,
                                   ssm_scenario: StateSpaceModel,
                                   x_previous: jnp.ndarray,
                                   t_previous: float,
                                   y_new: jnp.ndarray,
                                   t_new: float,
                                   random_keys: jnp.ndarray) -> jnp.ndarray:
        x_new = vmap(self.proposal_sample, (None, 0, None, None, None, 0))(ssm_scenario,
                                                                           x_previous, t_previous,
                                                                           y_new, t_new,
                                                                           random_keys)
        return x_new

    @partial(jit, static_argnums=(0, 1))
    def propose_and_intermediate_weight_vectorised(self,
                                                   ssm_scenario: StateSpaceModel,
                                                   x_previous: jnp.ndarray,
                                                   t_previous: float,
                                                   y_new: jnp.ndarray,
                                                   t_new: float,
                                                   random_keys: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:
        x_new = vmap(self.proposal_sample, (None, 0, None, None, None, 0))(ssm_scenario,
                                                                           x_previous, t_previous,
                                                                           y_new, t_new,
                                                                           random_keys)
        log_weight_new = vmap(self.intermediate_log_weight, (None, 0, None, 0, None, None))(ssm_scenario,
                                                                                            x_previous, t_previous,
                                                                                            x_new, y_new, t_new)
        return x_new, log_weight_new


class BootstrapFilter(ParticleFilter):
    name = 'Bootstrap Filter'

    def proposal_potential(self,
                           ssm_scenario: StateSpaceModel,
                           x_previous: jnp.ndarray,
                           t_previous: float,
                           x_new: jnp.ndarray,
                           y_new: jnp.ndarray,
                           t_new: float) -> Union[float, jnp.ndarray]:
        return ssm_scenario.transition_potential(x_previous, t_previous, x_new, t_new)

    def proposal_sample(self,
                        ssm_scenario: StateSpaceModel,
                        x_previous: jnp.ndarray,
                        t_previous: float,
                        y_new: jnp.ndarray,
                        t_new: float,
                        random_key: jnp.ndarray) -> jnp.ndarray:
        return ssm_scenario.transition_sample(x_previous, t_previous, t_new, random_key)

    def intermediate_log_weight(self,
                                ssm_scenario: StateSpaceModel,
                                x_previous: jnp.ndarray,
                                t_previous: float,
                                x_new: jnp.ndarray,
                                y_new: jnp.ndarray,
                                t_new: float) -> Union[float, jnp.ndarray]:
        return -ssm_scenario.likelihood_potential(x_new, y_new, t_new)


def initiate_particles(ssm_scenario: StateSpaceModel,
                       particle_filter: ParticleFilter,
                       n: int,
                       random_key: jnp.ndarray,
                       y: jnp.ndarray = None,
                       t: float = None) -> cdict:
    particle_filter.startup(ssm_scenario)

    sub_keys = random.split(random_key, n)

    init_vals, init_log_weight = particle_filter.initial_sample_and_weight_vectorised(ssm_scenario, y, t, sub_keys)

    if init_vals.ndim == 1:
        init_vals = init_vals[..., jnp.newaxis]

    initial_sample = cdict(value=init_vals[jnp.newaxis],
                           log_weight=init_log_weight[jnp.newaxis],
                           t=jnp.atleast_1d(t) if t is not None else jnp.zeros(1),
                           y=y[jnp.newaxis] if y is not None else None,
                           ess=jnp.atleast_1d(ess_log_weight(init_log_weight)))
    return initial_sample


def _resample(x_w_r: Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]:
    x, log_weight, random_key = x_w_r
    n = x.shape[-2]
    random_key, sub_key = random.split(random_key)
    return x[..., random.categorical(sub_key, log_weight, shape=(n,)), :], jnp.zeros(n), random_key


def resample_particles(particles: cdict,
                       random_key: jnp.ndarray,
                       resample_full: bool = True) -> cdict:
    out_particles = particles.copy()
    if out_particles.log_weight.ndim == 1:
        out_particles.value, out_particles.log_weight, _ = _resample((particles.value,
                                                                      particles.log_weight,
                                                                      random_key))
    elif resample_full:
        out_particles.value, latest_log_weight, _ = _resample((particles.value,
                                                               particles.log_weight[-1],
                                                               random_key))
        out_particles.log_weight = index_update(out_particles.log_weight, -1, latest_log_weight)
    else:
        latest_value, latest_log_weight, _ = _resample((particles.value[-1],
                                                        particles.log_weight[-1],
                                                        random_key))
        out_particles.value = index_update(out_particles.value, -1, latest_value)
        out_particles.log_weight = index_update(out_particles.log_weight, -1, latest_log_weight)
    return out_particles


def propagate_particle_filter(ssm_scenario: StateSpaceModel,
                              particle_filter: ParticleFilter,
                              particles: cdict,
                              y_new: jnp.ndarray,
                              t_new: float,
                              random_key: jnp.ndarray,
                              ess_threshold: float = 0.5,
                              resample_full: bool = True) -> cdict:
    n = particles.value.shape[1]
    ess_previous = particles.ess[-1]
    out_particles = cond(ess_previous < ess_threshold * n,
                         lambda tup: resample_particles(*tup),
                         lambda tup: tup[0].copy(),
                         (particles, random_key, resample_full))

    x_previous = out_particles.value[-1]
    log_weight_previous = out_particles.log_weight[-1]
    t_previous = out_particles.t[-1]

    split_keys = random.split(random_key, len(x_previous))

    x_new, log_weight_new = particle_filter.propose_and_intermediate_weight_vectorised(ssm_scenario,
                                                                                       x_previous, t_previous,
                                                                                       y_new, t_new, split_keys)

    log_weight_new = log_weight_previous + log_weight_new

    out_particles.value = jnp.append(out_particles.value, x_new, axis=0)
    out_particles.log_weight = jnp.append(out_particles.log_weight, log_weight_new, axis=0)
    out_particles.y = jnp.append(out_particles.y, y_new)
    out_particles.t = jnp.append(out_particles.t, t_new)
    out_particles.ess = jnp.append(out_particles.ess, ess_log_weight(log_weight_new))
    return out_particles


def run_particle_filter_for_marginals(ssm_scenario: StateSpaceModel,
                                      particle_filter: ParticleFilter,
                                      y: jnp.ndarray,
                                      t: jnp.ndarray,
                                      random_key: jnp.ndarray,
                                      n: int = None,
                                      initial_sample: cdict = None,
                                      ess_threshold: float = 0.5) -> cdict:
    if y.ndim == 1:
        y = y[..., jnp.newaxis]

    if initial_sample is None:
        random_key, sub_key = random.split(random_key)
        init_y = y[0]
        y = y[1:]

        initial_sample = initiate_particles(ssm_scenario, particle_filter, n, sub_key, init_y, t[0])
        t = t[1:]

    if n is None:
        n = initial_sample.value.shape[1]

    num_propagate_steps = len(y)
    int_rand_keys = random.split(random_key, num_propagate_steps)

    def particle_filter_body(samps_previous: cdict,
                             iter_ind: int) -> Tuple[cdict, cdict]:
        x_previous = samps_previous.value
        log_weight_previous = samps_previous.log_weight
        int_rand_key = int_rand_keys[iter_ind]

        ess_previous = samps_previous.ess

        x_res, log_weight_res, int_rand_key = cond(ess_previous < (ess_threshold * n),
                                                   _resample,
                                                   lambda _: (x_previous, log_weight_previous, int_rand_key),
                                                   (x_previous, log_weight_previous, int_rand_key))

        split_keys = random.split(int_rand_key, len(x_previous))

        x_new, log_weight_new = particle_filter.propose_and_intermediate_weight_vectorised(ssm_scenario,
                                                                                           x_res,
                                                                                           samps_previous.t,
                                                                                           y[iter_ind],
                                                                                           t[iter_ind],
                                                                                           split_keys)

        log_weight_new = log_weight_res + log_weight_new

        samps_new = samps_previous.copy()
        samps_new.value = x_new
        samps_new.log_weight = log_weight_new
        samps_new.y = y[iter_ind]
        samps_new.t = t[iter_ind]
        samps_new.ess = ess_log_weight(log_weight_new)
        return samps_new, samps_new

    _, after_init_samps = scan(particle_filter_body,
                               initial_sample[-1],
                               jnp.arange(num_propagate_steps))

    out_samps = initial_sample.copy()
    out_samps.value = jnp.append(initial_sample.value, after_init_samps.value, axis=0)
    out_samps.log_weight = jnp.append(initial_sample.log_weight, after_init_samps.log_weight, axis=0)
    out_samps.y = jnp.append(initial_sample.y, after_init_samps.y, axis=0)
    out_samps.t = jnp.append(initial_sample.t, after_init_samps.t)
    out_samps.ess = jnp.append(initial_sample.ess, after_init_samps.ess)

    return out_samps



########################################################################################################################
# Module: ssm/tests/test_ssm_linear_gaussian.py
# Description: Tests for linear TemporalGaussian state space models (and Kalman filtering/smoothing).
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

import unittest

import jax.numpy as jnp
from mocat.src.tests.test_ssm import TestSSM
from mocat.src.ssm.linear_gaussian.linear_gaussian import TimeHomogenousLinearGaussian


def create_eye_lgssm(dim: int):
    return TimeHomogenousLinearGaussian(initial_mean=jnp.zeros(dim),
                                        initial_covariance=jnp.eye(dim),
                                        transition_matrix=jnp.eye(dim),
                                        transition_covariance=jnp.eye(dim),
                                        likelihood_matrix=jnp.eye(dim),
                                        likelihood_covariance=jnp.eye(dim),
                                        name=f'{dim}D-eye LGSSM')


class Test1D(TestSSM):

    ssm_scenario = create_eye_lgssm(1)

    def test_simulate(self):
        super()._test_simulate()

    def test_bootstrap(self):
        super()._test_bootstrap()

    def test_backward(self):
        super()._test_backward()


class Test5D(TestSSM):

    ssm_scenario = create_eye_lgssm(5)

    def test_simulate(self):
        super()._test_simulate()

    def test_bootstrap(self):
        super()._test_bootstrap()

    def test_backward(self):
        super()._test_backward()


if __name__ == '__main__':
    unittest.main()



########################################################################################################################
# Module: utils.py
# Description: Some useful functions including TemporalGaussian potential evaluations and leapfrog integrator.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from typing import Any, Union, Callable, Tuple
from functools import partial
from warnings import warn

from jax.lax import scan, while_loop, cond
from jax import jit, numpy as jnp, vmap

from mocat.src.core import cdict


@partial(jit, static_argnums=(0,))
def leave_one_out_indices(n: int,
                          remove_index: int) -> jnp.ndarray:
    full_inds = jnp.arange(n + 1)
    leave_one_out_inds = jnp.where(jnp.arange(n) < remove_index, full_inds[:-1], full_inds[1:])
    return leave_one_out_inds


@jit
def _vectorised_gaussian_potential(x: jnp.ndarray,
                                   mean: Union[float, jnp.ndarray],
                                   sqrt_prec: jnp.ndarray) -> Union[float, jnp.ndarray]:
    diff = x - mean
    return 0.5 * jnp.sum(jnp.square(diff @ sqrt_prec), axis=-1)


@jit
def _mv_gaussian_potential(x: jnp.ndarray,
                           mean: Union[jnp.ndarray, float],
                           prec: jnp.ndarray) -> float:
    diff = x - mean
    return 0.5 * diff.T @ prec @ diff


@jit
def _mv_gaussian_potential_diag(x: jnp.ndarray,
                                mean: Union[jnp.ndarray, float],
                                prec_diag: Union[float, jnp.ndarray]) -> Union[float, jnp.ndarray]:
    diff = x - mean
    return 0.5 * jnp.sum(jnp.square(diff) * prec_diag, axis=-1)


def gaussian_potential(x: jnp.ndarray,
                       mean: Union[float, jnp.ndarray] = 0.,
                       prec: Union[float, jnp.ndarray] = None,
                       sqrt_prec: Union[float, jnp.ndarray] = None,
                       det_prec: float = None) -> Union[float, jnp.ndarray]:
    d = x.shape[-1]

    if prec is None and sqrt_prec is None:
        prec = 1.

    if isinstance(prec, float):
        prec = jnp.ones(d) * prec

    if isinstance(sqrt_prec, float):
        sqrt_prec = jnp.ones(d) * sqrt_prec

    if det_prec is None:
        if prec is not None and prec.ndim < 2:
            det_prec = jnp.prod(prec)
        elif sqrt_prec is not None and sqrt_prec.ndim < 2:
            det_prec = jnp.prod(sqrt_prec) ** 2

    if det_prec is None:
        # full precision matrix given but no det - computing without norm constant
        neg_log_z = 0
        warn('gaussian_potential queried with non-diagonal prec (or sqrt-prec) but no det_prec given'
             ' -> executing without normalising constant term')
    else:
        neg_log_z = (d * jnp.log(2 * jnp.pi) - jnp.log(det_prec)) / 2

    if x.ndim == 1 and sqrt_prec is None:
        # Single vals value (not vectorised)
        if prec is None:
            out_val = _mv_gaussian_potential_diag(x, mean, 1.)
        elif prec.ndim < 2:
            out_val = _mv_gaussian_potential_diag(x, mean, prec)
        else:
            out_val = _mv_gaussian_potential(x, mean, prec)
    else:
        # Multiple vals values (vectorised)
        if prec is not None and sqrt_prec is None:
            if prec.ndim < 2:
                sqrt_prec = jnp.sqrt(prec)
            else:
                sqrt_prec = jnp.linalg.cholesky(prec)
                warn('vectorised gaussian_potential queried with prec rather than sqrt_prec'
                     '-> executing using Cholesky decomp')

        if sqrt_prec is None:
            out_val = _mv_gaussian_potential_diag(x, mean, 1.)
        elif sqrt_prec.ndim < 2:
            out_val = _mv_gaussian_potential_diag(x, mean, sqrt_prec ** 2)
        else:
            out_val = _vectorised_gaussian_potential(x, mean, sqrt_prec)
    return out_val + neg_log_z


@partial(jit, static_argnums=(0,))
def leapfrog(potential_and_grad: Callable,
             state: cdict,
             stepsize: float,
             random_keys: jnp.ndarray) -> cdict:
    leapfrog_steps = len(random_keys)

    def leapfrog_step(init_state: cdict,
                      i: int):
        new_state = init_state.copy()

        p_half = init_state.momenta - stepsize / 2. * init_state.grad_potential
        new_state.value = init_state.value + stepsize * p_half

        new_state.potential, new_state.grad_potential = potential_and_grad(new_state.value, random_keys[i])

        new_state.momenta = p_half - stepsize / 2. * new_state.grad_potential

        next_sample_chain = new_state.copy()
        next_sample_chain.momenta = jnp.vstack([p_half, new_state.momenta])
        return new_state, next_sample_chain

    final_leapfrog, all_leapfrog = scan(leapfrog_step, state, jnp.arange(leapfrog_steps))

    all_leapfrog.momenta = jnp.concatenate(all_leapfrog.momenta)

    all_leapfrog = state[jnp.newaxis] + all_leapfrog

    return all_leapfrog


@partial(jit, static_argnums=(0, 1, 3))
def _while_loop_stacked(cond_fun: Callable,
                        body_fun: Callable,
                        init_carry: Tuple[Any, Any],
                        max_iter: int) -> Tuple[Any, int]:
    # init_carry = (init_state, extra)
    # cond_func: (state, extra) -> bool
    # body_func: (state, extra) -> (state, extra)
    # while_loop_stacked -> stacked(state)

    def update_cond(previous_carry: Any,
                    final_iter: int,
                    live_iter: int) -> bool:
        return cond(live_iter <= (final_iter + 1),
                    lambda args: cond_fun(*args),
                    lambda _: False,
                    previous_carry)

    def update_kernel(previous_val_and_final_iter: Tuple[Any, int],
                      live_iter: int) -> Tuple[Tuple[Any, int], Any]:
        previous_val, final_iter = previous_val_and_final_iter

        update_bool = update_cond(previous_val, final_iter, live_iter)

        final_iter = jnp.where(update_bool, live_iter, final_iter)

        new_val = cond(update_bool,
                       lambda args: body_fun(*args),
                       lambda _: init_carry,
                       previous_val)

        return (new_val, final_iter), new_val[0]

    final_val_and_final_iter, stack = scan(update_kernel, (init_carry, 0), jnp.arange(1, max_iter + 1))
    _, final_final_iter = final_val_and_final_iter
    return stack, final_final_iter


def while_loop_stacked(cond_fun: Callable,
                       body_fun: Callable,
                       init_carry: Tuple[Any, Any],
                       max_iter: int = 1000) -> Any:
    full_stack, final_int = _while_loop_stacked(cond_fun, body_fun, init_carry, max_iter)
    return tuple(a[:final_int] for a in full_stack) if isinstance(full_stack, tuple) else full_stack[:final_int]


@partial(jit, static_argnums=(0,))
def bisect(fun: Callable,
           bounds: Union[list, jnp.ndarray],
           max_iter: int = 1000,
           tol: float = 1e-5) -> Tuple[jnp.ndarray, jnp.ndarray, int]:
    evals = jnp.array([fun(bounds[0]), fun(bounds[1])])
    increasing_bool = evals[1] > evals[0]

    def conv_check(int_state: Tuple[jnp.ndarray, jnp.ndarray, int]) -> bool:
        int_bounds, int_evals, iter_ind = int_state
        return ~jnp.any(jnp.array([jnp.min(jnp.abs(int_evals)) < tol,
                                 iter_ind >= max_iter,
                                 jnp.all(int_evals < 0),
                                 jnp.all(int_evals > 0)]))

    def body_func(int_state: Tuple[jnp.ndarray, jnp.ndarray, int]) -> Tuple[jnp.ndarray, jnp.ndarray, int]:
        int_bounds, int_evals, iter_ind = int_state

        new_pos = int_bounds[0] - int_evals[0] * (int_bounds[1] - int_bounds[0]) / (int_evals[1] - int_evals[0])
        new_eval = fun(new_pos)

        replace_upper = jnp.where(increasing_bool, new_eval > 0, new_eval < 0)

        out_bounds = jnp.where(replace_upper, jnp.array([int_bounds[0], new_pos]), jnp.array([new_pos, int_bounds[1]]))
        out_evals = jnp.where(replace_upper, jnp.array([int_evals[0], new_eval]), jnp.array([new_eval, int_evals[1]]))

        return out_bounds, out_evals, iter_ind + 1

    fin_bounds, fin_evals, fin_iter = while_loop(conv_check,
                                                 body_func,
                                                 (bounds, evals, 0))

    return fin_bounds, fin_evals, fin_iter


def extract_dimension(*args):
    for a in args:
        if isinstance(a, jnp.ndarray) and a.ndim > 0:
            return a.shape[-1]
    return None


def reset_covariance(obj: Any,
                     key: str,
                     value: Any):
    prec_key = key.replace('covariance', 'precision')
    if value.ndim < 2:
        sqrt_val = jnp.sqrt(value)
        setattr(obj, key + '_sqrt', sqrt_val)
        setattr(obj, prec_key + '_sqrt', 1 / sqrt_val)
        setattr(obj, prec_key + '_det', 1 / value)
    else:
        sqrt_mat = jnp.linalg.cholesky(value)
        setattr(obj, key + '_sqrt', sqrt_mat)
        setattr(obj, prec_key + '_sqrt', jnp.linalg.inv(sqrt_mat))
        setattr(obj, prec_key + '_det', 1 / jnp.linalg.det(value))


def bfgs_update(prev_sqrt: jnp.ndarray,
                prev_inv_sqrt: jnp.ndarray,
                val_diff: jnp.ndarray,
                grad_diff: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:
    sty = val_diff.T @ grad_diff

    prev_hess_val_diff = prev_sqrt @ prev_sqrt.T @ val_diff
    hess_val_ip = val_diff.T @ prev_hess_val_diff

    p = val_diff / sty
    # q = jnp.sqrt(sty / (prev_hess_val_diff.T @ grad_diff)) * prev_hess_val_diff - grad_diff
    q = jnp.sqrt(sty / hess_val_ip) * prev_hess_val_diff + grad_diff

    t = val_diff / hess_val_ip
    u = jnp.sqrt(hess_val_ip / sty) * grad_diff + prev_hess_val_diff

    dim = val_diff.size
    new_inv_sqrt = (jnp.eye(dim) - jnp.outer(p, q)) @ prev_inv_sqrt
    new_sqrt = (jnp.eye(dim) - jnp.outer(u, t)) @ prev_sqrt

    return new_sqrt, new_inv_sqrt


def _cond_bfgs_update(prev_sqrt: jnp.ndarray,
                      prev_inv_sqrt: jnp.ndarray,
                      val_diff: jnp.ndarray,
                      grad_diff: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:
    sty = val_diff.T @ grad_diff
    return cond(jnp.logical_or(sty <= 0, jnp.any(jnp.isnan(val_diff))),
                lambda _: (prev_sqrt, prev_inv_sqrt),
                lambda tup: bfgs_update(*tup),
                (prev_sqrt, prev_inv_sqrt, val_diff, grad_diff))


@jit
def bfgs(initial_sqrt: jnp.ndarray,
         initial_inv_sqrt: jnp.ndarray,
         vals: jnp.ndarray,
         grads: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:
    final_carry, _ = scan(
        lambda prev_sqrts, i: (_cond_bfgs_update(*prev_sqrts, vals[i] - vals[i - 1], grads[i] - grads[i - 1]),
                               None),
        (initial_sqrt, initial_inv_sqrt),
        jnp.arange(1, len(vals)))
    return final_carry


@jit
def l2_distance_matrix(vals: jnp.ndarray) -> jnp.ndarray:
    return vmap(lambda x: vmap(lambda y: jnp.sum(jnp.square(x - y)))(vals))(vals) ** 0.5


def _gc_close(z_over_rad):
    return -0.25 * z_over_rad ** 5 \
           + 0.5 * z_over_rad ** 4 \
           + 5 / 8 * z_over_rad ** 3 \
           - 5 / 3 * z_over_rad ** 2 \
           + 1


def _gc_mid(z_over_rad):
    return 1 / 12 * z_over_rad ** 5 \
           - 0.5 * z_over_rad ** 4 \
           + 5 / 8 * z_over_rad ** 3 \
           + 5 / 3 * z_over_rad ** 2 \
           - 5 * z_over_rad \
           + 4 \
           - 2 / 3 / z_over_rad


def _gc(z, radius):
    z_over_rad = jnp.abs(z) / radius
    out = jnp.zeros_like(z)
    out = jnp.where(z < radius, _gc_close(z_over_rad), out)
    out = jnp.where((z >= radius) * (z < 2 * radius), _gc_mid(z_over_rad), out)
    return out


def gc_matrix(dim: int,
              radius: float):
    return vmap(lambda i: _gc(jnp.min(jnp.array([jnp.abs(jnp.arange(dim) - i),
                                               jnp.abs(jnp.arange(dim) + dim - i),
                                               jnp.abs(jnp.arange(dim) - dim - i)]), axis=0), radius)
                )(jnp.arange(dim))


@jit
def kalman_gain(cov_sqrt: jnp.ndarray,
                lik_mat: jnp.ndarray,
                lik_cov_sqrt: jnp.ndarray,
                inv_mat: jnp.ndarray = None) -> jnp.ndarray:
    if inv_mat is None:
        lik_mat_cov_sqrt = lik_mat @ cov_sqrt
        inv_mat = jnp.linalg.inv(lik_mat_cov_sqrt @ lik_mat_cov_sqrt.T + lik_cov_sqrt @ lik_cov_sqrt.T)
    return cov_sqrt @ cov_sqrt.T @ lik_mat.T @ inv_mat


########################################################################################################################
# Module: transport/smc.py
# Description: SMC samplers for tempered sequence of distributions.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from typing import Tuple, Union, Type
from inspect import isclass

from jax import numpy as jnp, random, vmap
from jax.lax import scan, cond

from mocat.src.core import Scenario, cdict
from mocat.src.transport.sampler import TransportSampler
from mocat.src.mcmc.sampler import MCMCSampler, Correction, check_correction
from mocat.src.utils import bisect
from mocat.src.metrics import log_ess_log_weight


class SMCSampler(TransportSampler):
    name = 'SMC Sampler'

    def forward_proposal(self,
                         scenario: Scenario,
                         previous_state: cdict,
                         previous_extra: cdict) -> Tuple[cdict, cdict]:
        raise AttributeError(f'{self.name} forward_proposal not initiated')

    def adapt(self,
              ensemble_state: cdict,
              ensemble_extra: cdict) -> Tuple[cdict, cdict]:
        return ensemble_state, ensemble_extra

    def resample_criterion(self,
                           ensemble_state: cdict,
                           ensemble_extra: cdict) -> bool:
        return True

    def resample(self,
                 ensemble_state: cdict,
                 ensemble_extra: cdict) -> Tuple[cdict, cdict]:
        n = ensemble_state.value.shape[0]
        resampled_indices = random.categorical(ensemble_extra.random_key[0],
                                               ensemble_state.log_weight,
                                               shape=(n,))
        resampled_ensemble_state = ensemble_state[resampled_indices]
        resampled_ensemble_extra = ensemble_extra[resampled_indices]
        resampled_ensemble_state.log_weight = jnp.zeros(n)
        return resampled_ensemble_state, resampled_ensemble_extra

    def log_weight(self,
                   previous_ensemble_state: cdict,
                   previous_ensemble_extra: cdict,
                   new_ensemble_state: cdict,
                   new_ensemble_extra: cdict) -> jnp.ndarray:
        return previous_ensemble_state.log_weight + new_ensemble_state.log_weight

    def update(self,
               scenario: Scenario,
               ensemble_state: cdict,
               ensemble_extra: cdict) -> Tuple[cdict, cdict]:
        ensemble_extra.iter = ensemble_extra.iter + 1

        resample_bool = self.resample_criterion(ensemble_state, ensemble_extra)

        resampled_ensemble_state, resampled_ensemble_extra \
            = cond(resample_bool,
                   lambda tup: self.resample(*tup),
                   lambda tup: tup,
                   (ensemble_state, ensemble_extra))

        advanced_state, advanced_extra = vmap(self.forward_proposal, in_axes=(None, 0, 0))(scenario,
                                                                                           resampled_ensemble_state,
                                                                                           resampled_ensemble_extra)

        advanced_state.log_weight = self.log_weight(resampled_ensemble_state, resampled_ensemble_extra,
                                                    advanced_state, advanced_extra)

        advanced_state, advanced_extra = self.adapt(advanced_state, advanced_extra)

        return advanced_state, advanced_extra


class TemperedSMCSampler(SMCSampler):
    name = 'Tempered SMC Sampler'

    def __init__(self,
                 temperature_schedule: Union[None, jnp.ndarray] = None,
                 max_temperature: float = 1.,
                 max_iter: int = int(1e4),
                 **kwargs):
        self.max_iter = max_iter
        self.max_temperature = max_temperature
        self.temperature_schedule = temperature_schedule
        super().__init__(**kwargs)

    def __setattr__(self, key, value):
        super().__setattr__(key, value)
        if key == 'temperature_schedule':
            if value is None:
                self.next_temperature = self.next_temperature_adaptive
                self.max_iter = int(1e4)
            else:
                self.next_temperature = lambda state, extra: self.temperature_schedule[extra.iter[0]]
                self.max_temperature = value[-1]
                self.max_iter = len(value)

    def startup(self,
                scenario: Scenario,
                n: int,
                initial_state: cdict,
                initial_extra: cdict,
                **kwargs) -> Tuple[cdict, cdict]:

        initial_state, initial_extra = super().startup(scenario, n, initial_state, initial_extra, **kwargs)

        random_keys = random.split(initial_extra.random_key[0], 3*n)

        initial_extra.random_key = random_keys[:n]

        n = len(initial_state.value)
        initial_state.prior_potential = vmap(scenario.prior_potential)(initial_state.value, random_keys[n:(2*n)])
        initial_state.likelihood_potential = vmap(scenario.likelihood_potential)(initial_state.value,
                                                                                 random_keys[(2*n):])
        initial_state.potential = initial_state.prior_potential
        initial_state.temperature = jnp.zeros(n)
        initial_state.log_weight = jnp.zeros(n)

        scenario.temperature = 0.

        initial_extra.parameters = vmap(lambda _: initial_extra.parameters)(jnp.arange(n))

        return initial_state, initial_extra

    def next_temperature_adaptive(self,
                                  ensemble_state: cdict,
                                  ensemble_extra: cdict) -> float:
        raise AttributeError(f'{self.name} next_temperature_adaptive not initiated')

    def termination_criterion(self,
                              ensemble_state: cdict,
                              ensemble_extra: cdict) -> bool:
        return jnp.logical_or(ensemble_state.temperature[0] >= self.max_temperature,
                             ensemble_extra.iter[0] >= self.max_iter)

    def clean_chain(self,
                    scenario: Scenario,
                    chain_ensemble_state: cdict) -> cdict:
        chain_ensemble_state.temperature = chain_ensemble_state.temperature[:, 0]
        scenario.temperature = float(chain_ensemble_state.temperature[-1])
        return chain_ensemble_state

    def update(self,
               scenario: Scenario,
               ensemble_state: cdict,
               ensemble_extra: cdict) -> Tuple[cdict, cdict]:
        n = ensemble_state.value.shape[0]
        ensemble_extra.iter = ensemble_extra.iter + 1
        prev_temperature = ensemble_state.temperature[0]
        scenario.temperature = prev_temperature

        resample_bool = self.resample_criterion(ensemble_state, ensemble_extra)
        ensemble_state.log_weight = jnp.where(resample_bool, 0., ensemble_state.log_weight)

        resampled_ensemble_state, resampled_ensemble_extra \
            = cond(resample_bool,
                   lambda tup: self.resample(*tup),
                   lambda tup: tup,
                   (ensemble_state, ensemble_extra))

        advanced_state, advanced_extra = vmap(self.forward_proposal, in_axes=(None, 0, 0))(scenario,
                                                                                           resampled_ensemble_state,
                                                                                           resampled_ensemble_extra)

        advanced_extra.iter = ensemble_extra.iter
        next_temperature = self.next_temperature(advanced_state, advanced_extra)
        advanced_state.temperature = jnp.ones(n) * next_temperature
        advanced_state.log_weight = self.log_weight(ensemble_state, ensemble_extra, advanced_state, advanced_extra)

        advanced_state, advanced_extra = self.adapt(advanced_state, advanced_extra)

        return advanced_state, advanced_extra


class MetropolisedSMCSampler(TemperedSMCSampler):
    name = "Metropolised SMC Sampler"

    def __init__(self,
                 mcmc_sampler: Union[MCMCSampler, Type[MCMCSampler]],
                 mcmc_correction: Union[Correction, Type[Correction], str] = 'sampler_default',
                 mcmc_steps: int = 20,
                 max_iter: int = int(1e4),
                 temperature_schedule: Union[None, jnp.ndarray] = None,
                 max_temperature: float = 1.,
                 ess_threshold: float = 0.8,
                 bisection_tol: float = 1e-5,
                 max_bisection_iter: int = 1000,
                 **kwargs):
        if temperature_schedule is not None:
            if temperature_schedule[0] == 0.:
                temperature_schedule = temperature_schedule[1:]

        super().__init__(max_iter=max_iter, temperature_schedule=temperature_schedule, max_temperature=max_temperature,
                         **kwargs)
        if isclass(mcmc_sampler):
            mcmc_sampler = mcmc_sampler()
        self.mcmc_sampler = mcmc_sampler
        if mcmc_correction != 'sampler_default':
            self.mcmc_sampler.correction = mcmc_correction
        self.parameters.mcmc_steps = mcmc_steps

        self.parameters.ess_threshold = ess_threshold
        self.parameters.bisection_tol = bisection_tol
        self.parameters.max_bisection_iter = max_bisection_iter

    def __setattr__(self, key, value):
        if key == 'temperature_schedule':
            if value is not None and value[0] == 0.:
                value = value[1:]
        super().__setattr__(key, value)

    def startup(self,
                scenario: Scenario,
                n: int,
                initial_state: cdict,
                initial_extra: cdict,
                **kwargs) -> Tuple[cdict, cdict]:

        self.mcmc_sampler.correction = check_correction(self.mcmc_sampler.correction)

        initial_state, initial_extra = super().startup(scenario, n, initial_state, initial_extra, **kwargs)

        first_temp = self.next_temperature(initial_state, initial_extra)
        initial_state.temperature += first_temp
        initial_state.potential = initial_state.prior_potential + first_temp * initial_state.likelihood_potential
        initial_state.log_weight = - first_temp * initial_state.likelihood_potential

        initial_state, initial_extra = vmap(
            lambda state, extra: self.mcmc_sampler.startup(scenario,
                                                           n,
                                                           state,
                                                           extra))(initial_state, initial_extra)

        return initial_state, initial_extra

    @staticmethod
    def log_ess(current_temperature: float,
                new_temperature: float,
                likelihood_potential: jnp.ndarray) -> float:
        log_weight = - (new_temperature - current_temperature) * likelihood_potential
        return log_ess_log_weight(log_weight)

    def next_temperature_adaptive(self,
                                  ensemble_state: cdict,
                                  ensemble_extra: cdict) -> float:
        temperature_bounds = jnp.array([ensemble_state.temperature[0], self.max_temperature])
        likelihood_potential = ensemble_state.likelihood_potential
        log_n_samp_threshold = jnp.log(len(likelihood_potential) * self.parameters.ess_threshold)

        bisect_out_bounds, bisect_out_evals, bisect_out_iter \
            = bisect(lambda x: self.log_ess(temperature_bounds[0],
                                            x,
                                            likelihood_potential) - log_n_samp_threshold,
                     temperature_bounds,
                     max_iter=self.parameters.max_bisection_iter,
                     tol=self.parameters.bisection_tol)
        return bisect_out_bounds[jnp.argmin(jnp.abs(bisect_out_evals))]

    def clean_mcmc_chain(self,
                         chain_state: cdict,
                         chain_extra: cdict) -> Tuple[cdict, cdict]:
        clean_state = chain_state[-1]
        clean_extra = chain_extra[-1]
        clean_state.alpha = chain_state.alpha.mean()
        clean_extra.parameters = chain_extra.parameters[-1]
        return clean_state, clean_extra

    def forward_proposal(self,
                         scenario: Scenario,
                         state: cdict,
                         extra: cdict) -> Tuple[cdict, cdict]:

        def mcmc_kernel(previous_carry: Tuple[cdict, cdict],
                        _: None) -> Tuple[Tuple[cdict, cdict], Tuple[cdict, cdict]]:
            new_carry = self.mcmc_sampler.update(scenario, *previous_carry)
            return new_carry, new_carry

        start_state, start_extra = self.mcmc_sampler.startup(scenario,
                                                             state.value.shape[0],
                                                             state,
                                                             extra)

        final_carry, chain = scan(mcmc_kernel,
                                  (start_state, start_extra),
                                  None,
                                  length=self.parameters.mcmc_steps)

        advanced_state, advanced_extra = self.clean_mcmc_chain(chain[0], chain[1])

        advanced_extra.random_key, subkey = random.split(advanced_extra.random_key)

        advanced_state.prior_potential = scenario.prior_potential(advanced_state.value, subkey)
        advanced_state.likelihood_potential = (advanced_state.potential - advanced_state.prior_potential) \
                                              / scenario.temperature
        advanced_extra.iter = extra.iter
        return advanced_state, advanced_extra

    def log_weight(self,
                   previous_ensemble_state: cdict,
                   previous_ensemble_extra: cdict,
                   new_ensemble_state: cdict,
                   new_ensemble_extra: cdict) -> jnp.ndarray:
        return - (new_ensemble_state.temperature[0] - previous_ensemble_state.temperature[0]) \
               * new_ensemble_state.likelihood_potential





from mocat.src.utils import leapfrog
from mocat.src.utils import gaussian_potential
from mocat.src.utils import leave_one_out_indices
from mocat.src.utils import while_loop_stacked
from mocat.src.utils import bisect
from mocat.src.utils import extract_dimension
from mocat.src.utils import reset_covariance
from mocat.src.utils import bfgs_update
from mocat.src.utils import bfgs

from mocat.src.core import impl_checkable
from mocat.src.core import is_implemented

